@@ -72,6 +72,8 @@ def dbg(debug, *args, **kwargs):
 
 def bstr(buf):
   """Decode the byte string into a string"""
+  if isinstance(buf, str):
+      return buf
   return buf.decode('utf-8')
 
 
@@ -162,6 +164,7 @@ basedir=${base%%/*}
 exec \
   "${basedir}/%(interp)s" \
   --library-path "%(libpaths)s" \
+  --inhibit-cache \
   --inhibit-rpath '' \
   "${base}.elf" \
   "$@"
