@@ -83,6 +83,10 @@
 #ifdef __SANITIZE_ADDRESS__
 # define PAX_UTILS_CLEANUP 1
 #endif
+/* Coverity catches some things we leak on purpose. */
+#ifdef __COVERITY__
+# define PAX_UTILS_CLEANUP 1
+#endif
 #ifndef PAX_UTILS_CLEANUP
 # define PAX_UTILS_CLEANUP 0
 #endif
@@ -217,6 +221,9 @@
 #ifndef O_CLOEXEC
 # define O_CLOEXEC 0
 #endif
+#ifndef O_PATH
+# define O_PATH 0
+#endif
 
 #define __unused__ __attribute__((__unused__))
 
